@@ -1,6 +1,6 @@
 /* 
    HTTP Authentication routines
-   Copyright (C) 1999-2002, Joe Orton <joe@manyfish.co.uk>
+   Copyright (C) 1999-2003, Joe Orton <joe@manyfish.co.uk>
 
    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Library General Public
@@ -32,18 +32,14 @@
 #ifdef HAVE_STDLIB_H
 #include <stdlib.h>
 #endif
-
 #ifdef HAVE_STRING_H
 #include <string.h>
 #endif
-
-#include <time.h>
-
-#ifdef HAVE_SNPRINTF_H
-#include "snprintf.h"
+#ifdef HAVE_STRINGS_H
+#include <strings.h>
 #endif
 
-#include "base64.h"
+#include <time.h>
 
 #include "ne_md5.h"
 #include "ne_dates.h"
@@ -85,7 +81,6 @@ typedef enum {
 struct auth_challenge {
     auth_scheme scheme;
     const char *realm;
-    const char *domain;
     const char *nonce;
     const char *opaque;
     unsigned int stale:1; /* if stale=true */
@@ -107,23 +102,27 @@ static const char *algorithm_names[] = {
     NULL
 };
 
-/* The callback used to request the username and password in the given
- * realm. The username and password must be placed in malloc()-allocate
- * memory.
- * Must return:
- *   0 on success, 
- *  -1 to cancel.
- */
+static const struct auth_class {
+    const char *id, *req_hdr, *resp_hdr, *resp_info_hdr, *fail_msg;
+    int status_code, fail_code;
+} ah_server_class = {
+    HOOK_SERVER_ID,
+    "Authorization", "WWW-Authenticate", "Authentication-Info",
+    N_("Server was not authenticated correctly."), 401, NE_AUTH
+}, ah_proxy_class = {
+    HOOK_PROXY_ID,
+    "Proxy-Authorization", "Proxy-Authenticate", "Proxy-Authentication-Info",
+    N_("Proxy server was not authenticated correctly."), 407, NE_PROXYAUTH 
+};
 
 /* Authentication session state. */
 typedef struct {
     ne_session *sess;
-
-    /* for making this proxy/server generic. */
-    const char *req_hdr, *resp_hdr, *resp_info_hdr, *fail_msg;
-    int status_code;
-    int fail_code;
-
+    
+    /* Specifics for server/proxy auth. FIXME: need a better field
+     * name! */
+    const struct auth_class *spec;
+    
     /* The scheme used for this authentication session */
     auth_scheme scheme;
     /* The callback used to request new username+password */
@@ -143,9 +142,6 @@ typedef struct {
     char *nonce;
     char *cnonce;
     char *opaque;
-    /* A list of domain strings */
-    unsigned int domain_count;
-    char **domain;
     auth_qop qop;
     auth_algorithm alg;
     int nonce_count;
@@ -163,9 +159,6 @@ typedef struct {
     unsigned int port;
 
     int attempt;
-
-    /* the current request structure */
-    struct auth_request *req;
 } auth_session;
 
 struct auth_request {
@@ -185,60 +178,6 @@ struct auth_request {
     char *auth_hdr, *auth_info_hdr;
 };
 
-/* Private prototypes which used to be public. */
-
-static auth_session *auth_create(ne_session *sess,
-				 ne_request_auth callback, 
-				 void *userdata);
-
-/* Pass this the value of the "(Proxy,WWW)-Authenticate: " header field.
- * Returns:
- *   0 if we can now authenticate ourselves with the server.
- *   non-zero if we can't
- */
-static int auth_challenge(auth_session *sess, const char *value);
-
-/* If you receive a "(Proxy-)Authentication-Info:" header, pass its value to
- * this function. Returns zero if this successfully authenticates
- * the response as coming from the server, and false if it hasn't. */
-static int verify_response(struct auth_request *req, 
-			   auth_session *sess, const char *value);
-
-/* Private prototypes */
-static char *get_cnonce(void);
-static void clean_session(auth_session *sess);
-static int digest_challenge(auth_session *, struct auth_challenge *);
-static int basic_challenge(auth_session *, struct auth_challenge *);
-static char *request_digest(auth_session *sess, struct auth_request *req);
-static char *request_basic(auth_session *);
-
-/* Domain handling */
-static int is_in_domain(auth_session *sess, const char *uri);
-static int parse_domain(auth_session *sess, const char *domain);
-
-auth_session *auth_create(ne_session *sess,
-			  ne_request_auth callback, 
-			  void *userdata) 
-{
-    auth_session *ret = ne_calloc(sizeof(auth_session));
-
-    ret->reqcreds = callback;
-    ret->reqcreds_udata = userdata;
-    ret->sess = sess;
-   
-    return ret;
-}
-
-#if 0
-void ne_auth_set_server(auth_session *sess, 
-			   const char *host, unsigned int port, const char *scheme)
-{
-    sess->host = host;
-    sess->port = port;
-    sess->req_scheme = scheme;
-}
-#endif
-
 static void clean_session(auth_session *sess) 
 {
     sess->can_handle = 0;
@@ -247,10 +186,6 @@ static void clean_session(auth_session *sess)
     NE_FREE(sess->cnonce);
     NE_FREE(sess->opaque);
     NE_FREE(sess->realm);
-    if (sess->domain_count > 0) {
-	split_string_free(sess->domain);
-	sess->domain_count = 0;
-    }
 }
 
 /* Returns cnonce-value. We just use base64(time).
@@ -270,56 +205,6 @@ static int get_credentials(auth_session *sess, char *pwbuf)
 			    sess->username, pwbuf);
 }
 
-static int parse_domain(auth_session *sess, const char *domain) {
-    char **doms;
-    int count, ret;
-
-    doms = split_string_c(domain, ' ', NULL, " \r\n\t", &count);
-    if (doms != NULL) {
-	if (count > 0) {
-	    sess->domain = doms;
-	    sess->domain_count = count;
-	    ret = 0;
-	} else {
-	    free(doms);
-	    ret = -1;
-	}
-    } else {
-	ret = -1;
-    }
-
-    return ret;
-}
-
-/* Returns:
- *    0: if uri is in NOT in domain of session
- * else: uri IS in domain of session (or no domain known)
- */
-static int is_in_domain(auth_session *sess, const char *uri)
-{
-#if 1
-    return 1;
-#else
-    /* TODO: Need proper URI comparison for this to work. */
-    int ret, dom;
-    const char *abs_path;
-    if (sess->domain_count == 0) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "No domain, presuming okay.\n");
-	return 1;
-    }
-    ret = 0;
-    abs_path = uri_abspath(uri);
-    for (dom = 0; dom < sess->domain_count; dom++) {
-	NE_DEBUG(NE_DBG_HTTPAUTH, "Checking domain: %s\n", sess->domain[dom]);
-	if (uri_childof(sess->domain[dom], abs_path)) {
-	    ret = 1;
-	    break;
-	}
-    }
-    return ret;
-#endif
-}
-
 /* Examine a Basic auth challenge.
  * Returns 0 if an valid challenge, else non-zero. */
 static int 
@@ -346,12 +231,12 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
 
     sess->scheme = auth_scheme_basic;
 
-    CONCAT3(tmp, sess->username, ":", password);
+    tmp = ne_concat(sess->username, ":", password, NULL);
     sess->basic = ne_base64(tmp, strlen(tmp));
     free(tmp);
 
     /* Paranoia. */
-    memset(password, 0, NE_ABUFSIZ);
+    memset(password, 0, sizeof password);
 
     return 0;
 }
@@ -359,9 +244,7 @@ basic_challenge(auth_session *sess, struct auth_challenge *parms)
 /* Add Basic authentication credentials to a request */
 static char *request_basic(auth_session *sess) 
 {
-    char *buf;
-    CONCAT3(buf, "Basic ", sess->basic, "\r\n");
-    return buf;
+    return ne_concat("Basic ", sess->basic, "\r\n", NULL);
 }
 
 /* Examine a digest challenge: return 0 if it is a valid Digest challenge,
@@ -403,14 +286,7 @@ static int digest_challenge(auth_session *sess,
     sess->scheme = auth_scheme_digest;
     sess->nonce = ne_strdup(parms->nonce);
     sess->cnonce = get_cnonce();
-    if (parms->domain) {
-	if (parse_domain(sess, parms->domain)) {
-	    /* TODO: Handle the error? */
-	}
-    } else {
-	sess->domain = NULL;
-	sess->domain_count = 0;
-    }
+    /* TODO: add domain handling. */
     if (parms->opaque != NULL) {
 	sess->opaque = ne_strdup(parms->opaque); /* don't strip the quotes */
     }
@@ -440,7 +316,7 @@ static int digest_challenge(auth_session *sess,
 	ne_md5_process_bytes(sess->realm, strlen(sess->realm), &tmp);
 	ne_md5_process_bytes(":", 1, &tmp);
 	ne_md5_process_bytes(password, strlen(password), &tmp);
-	memset(password, 0, NE_ABUFSIZ); /* done with that. */
+	memset(password, 0, sizeof password); /* done with that. */
 	ne_md5_finish_ctx(&tmp, tmp_md5);
 	if (sess->alg == auth_alg_md5_sess) {
 	    unsigned char a1_md5[16];
@@ -660,7 +536,8 @@ static int tokenize(char **hdr, char **key, char **value,
  *    0 if it gives a valid authentication for the server 
  *    non-zero otherwise (don't believe the response in this case!).
  */
-int verify_response(struct auth_request *req, auth_session *sess, const char *value) 
+static int verify_response(struct auth_request *req, auth_session *sess,
+			   const char *value) 
 {
     char *hdr, *pnt, *key, *val;
     auth_qop qop = auth_qop_none;
@@ -800,9 +677,10 @@ int verify_response(struct auth_request *req, auth_session *sess, const char *va
     return okay;
 }
 
-
-/* A new challenge presented by the server */
-int auth_challenge(auth_session *sess, const char *value) 
+/* Passed the value of a "(Proxy,WWW)-Authenticate: " header field.
+ * Returns 0 if valid challenge was accepted; non-zero if no valid
+ * challenge was found. */
+static int auth_challenge(auth_session *sess, const char *value) 
 {
     char *pnt, *key, *val, *hdr;
     struct auth_challenge *chall = NULL, *challenges = NULL;
@@ -858,8 +736,6 @@ int auth_challenge(auth_session *sess, const char *value)
 	    chall->nonce = val;
 	} else if (strcasecmp(key, "opaque") == 0) {
 	    chall->opaque = val;
-	} else if (strcasecmp(key, "domain") == 0) {
-	    chall->domain = val;
 	} else if (strcasecmp(key, "stale") == 0) {
 	    /* Truth value */
 	    chall->stale = (strcasecmp(val, "true") == 0);
@@ -871,7 +747,7 @@ int auth_challenge(auth_session *sess, const char *value)
 	    } else {
 		chall->alg = auth_alg_unknown;
 	    }
-	} else if (strcasecmp(val, "qop") == 0) {
+	} else if (strcasecmp(key, "qop") == 0) {
 	    char **qops;
 	    int qop;
 	    qops = split_string(val, ',', NULL, " \r\n\t");
@@ -883,6 +759,7 @@ int auth_challenge(auth_session *sess, const char *value)
 		    chall->qop_auth_int = 1;
 		}
 	    }
+	    split_string_free(qops);
 	}
     }
 
@@ -949,43 +826,39 @@ static void auth_body_reader(void *cookie, const char *block, size_t length)
     ne_md5_process_bytes(block, length, ctx);
 }
 
-static void ah_create(void *session, ne_request *request, const char *method,
+static void ah_create(ne_request *req, void *session, const char *method,
 		      const char *uri)
 {
     auth_session *sess = session;
     struct auth_request *areq = ne_calloc(sizeof *areq);
 
-    NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->resp_hdr);
+    NE_DEBUG(NE_DBG_HTTPAUTH, "ah_create, for %s\n", sess->spec->resp_hdr);
 
     areq->method = method;
     areq->uri = uri;
-    areq->request = request;
+    areq->request = req;
 
-    ne_add_response_header_handler(request, sess->resp_hdr,
+    ne_add_response_header_handler(req, sess->spec->resp_hdr,
 				   ne_duplicate_header, &areq->auth_hdr);
 
 	
-    ne_add_response_header_handler(request, sess->resp_info_hdr,
+    ne_add_response_header_handler(req, sess->spec->resp_info_hdr,
 				   ne_duplicate_header, 
 				   &areq->auth_info_hdr);
     
     sess->attempt = 0;
-    sess->req = areq;
+
+    ne_set_request_private(req, sess->spec->id, areq);
 }
 
 
-static void ah_pre_send(void *cookie, ne_buffer *request)
+static void ah_pre_send(ne_request *r, void *cookie, ne_buffer *request)
 {
     auth_session *sess = cookie;
-    struct auth_request *req = sess->req;
+    struct auth_request *req = ne_get_request_private(r, sess->spec->id);
 
     if (!sess->can_handle) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Not handling session.\n");
-    } else if (!is_in_domain(sess, req->uri)) {
-	/* We have moved out of the authentication domain */
-	NE_DEBUG(NE_DBG_HTTPAUTH, 
-	      "URI %s outside session domain, not handling.\n", req->uri);
-	req->will_handle = 0;
     } else {
 	char *value;
 
@@ -1013,7 +886,7 @@ static void ah_pre_send(void *cookie, ne_buffer *request)
 	}
 
 	if (value != NULL) {
-	    ne_buffer_concat(request, sess->req_hdr, ": ", value, NULL);
+	    ne_buffer_concat(request, sess->spec->req_hdr, ": ", value, NULL);
 	    free(value);
 	}
 
@@ -1023,29 +896,29 @@ static void ah_pre_send(void *cookie, ne_buffer *request)
 
 #define SAFELY(x) ((x) != NULL?(x):"null")
 
-static int ah_post_send(void *cookie, const ne_status *status)
+static int ah_post_send(ne_request *req, void *cookie, const ne_status *status)
 {
     auth_session *sess = cookie;
-    struct auth_request *areq = sess->req;
+    struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
     int ret = NE_OK;
 
     NE_DEBUG(NE_DBG_HTTPAUTH, 
 	  "ah_post_send (#%d), code is %d (want %d), %s is %s\n",
-	  sess->attempt, status->code, sess->status_code, 
-	  SAFELY(sess->resp_hdr), SAFELY(areq->auth_hdr));
+	  sess->attempt, status->code, sess->spec->status_code, 
+	  sess->spec->resp_hdr, SAFELY(areq->auth_hdr));
     if (areq->auth_info_hdr != NULL && 
 	verify_response(areq, sess, areq->auth_info_hdr)) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Response authentication invalid.\n");
-	ne_set_error(sess->sess, sess->fail_msg);
+	ne_set_error(sess->sess, _(sess->spec->fail_msg));
 	ret = NE_ERROR;
-    } else if (status->code == sess->status_code && 
+    } else if (status->code == sess->spec->status_code && 
 	       areq->auth_hdr != NULL) {
 	NE_DEBUG(NE_DBG_HTTPAUTH, "Got challenge with code %d.\n", status->code);
 	if (!auth_challenge(sess, areq->auth_hdr)) {
 	    ret = NE_RETRY;
 	} else {
 	    clean_session(sess);
-	    ret = sess->fail_code;
+	    ret = sess->spec->fail_code;
 	}
     }
 
@@ -1055,10 +928,11 @@ static int ah_post_send(void *cookie, const ne_status *status)
     return ret;
 }
 
-static void ah_destroy(void *session)
+static void ah_destroy(ne_request *req, void *session)
 {
     auth_session *sess = session;
-    free(sess->req);
+    struct auth_request *areq = ne_get_request_private(req, sess->spec->id);
+    free(areq);
 }
 
 static void free_auth(void *cookie)
@@ -1069,52 +943,45 @@ static void free_auth(void *cookie)
     free(sess);
 }
 
-static void register_hooks(ne_session *sess, auth_session *as, const char *id)
+static void auth_register(ne_session *sess,
+			  const struct auth_class *ahc, const char *id, 
+			  ne_request_auth callback, void *userdata) 
 {
-    ne_hook_create_request(sess, ah_create, as);
-    ne_hook_pre_send(sess, ah_pre_send, as);
-    ne_hook_post_send(sess, ah_post_send, as);
-    ne_hook_destroy_request(sess, ah_destroy, as);
-    ne_hook_destroy_session(sess, free_auth, as);
+    auth_session *ahs = ne_calloc(sizeof *ahs);
 
-    ne_hook_session_accessor(sess, id, ne_null_accessor, as);
+    ahs->reqcreds = callback;
+    ahs->reqcreds_udata = userdata;
+    ahs->sess = sess;
+    ahs->spec = ahc;
+
+    /* Register hooks */
+    ne_hook_create_request(sess, ah_create, ahs);
+    ne_hook_pre_send(sess, ah_pre_send, ahs);
+    ne_hook_post_send(sess, ah_post_send, ahs);
+    ne_hook_destroy_request(sess, ah_destroy, ahs);
+    ne_hook_destroy_session(sess, free_auth, ahs);
+
+    ne_set_session_private(sess, id, ahs);
 }
 
 void ne_set_server_auth(ne_session *sess, ne_request_auth callback, 
-			  void *userdata)
+			void *userdata)
 {
-    auth_session *auth_sess = auth_create(sess, callback, userdata);
-
-    /* Server auth details */
-    auth_sess->status_code = 401;
-    auth_sess->fail_code = NE_AUTH;
-    auth_sess->resp_hdr = "WWW-Authenticate";
-    auth_sess->resp_info_hdr = "Authentication-Info";
-    auth_sess->req_hdr = "Authorization";
-    auth_sess->fail_msg = _("Server was not authenticated correctly.");
-
-    register_hooks(sess, auth_sess, HOOK_SERVER_ID);
+    auth_register(sess, &ah_server_class, HOOK_SERVER_ID, callback, userdata);
 }
 
 void ne_set_proxy_auth(ne_session *sess, ne_request_auth callback, 
-			 void *userdata)
+		       void *userdata)
 {
-    auth_session *auth_sess = auth_create(sess, callback, userdata);
-
-    /* Proxy auth details */
-    auth_sess->status_code = 407;
-    auth_sess->fail_code = NE_PROXYAUTH;
-    auth_sess->resp_hdr = "Proxy-Authenticate";
-    auth_sess->resp_info_hdr = "Proxy-Authentication-Info";
-    auth_sess->req_hdr = "Proxy-Authorization";
-    auth_sess->fail_msg = _("Proxy server was not authenticated correctly.");
-
-    register_hooks(sess, auth_sess, HOOK_PROXY_ID);
+    auth_register(sess, &ah_proxy_class, HOOK_PROXY_ID, callback, userdata);
 }
 
 void ne_forget_auth(ne_session *sess)
 {
-    clean_session(ne_session_hook_private(sess, HOOK_SERVER_ID));
-    clean_session(ne_session_hook_private(sess, HOOK_PROXY_ID));
+    auth_session *as;
+    if ((as = ne_get_session_private(sess, HOOK_SERVER_ID)) != NULL)
+	clean_session(as);
+    if ((as = ne_get_session_private(sess, HOOK_PROXY_ID)) != NULL)
+	clean_session(as);
 }
 
