@@ -33,10 +33,13 @@
 #include <errno.h>
 #endif
 
+#include <ctype.h> /* for cleanse() */
+
 #ifdef NEON_SSL
 #include <openssl/ssl.h>
 #include <openssl/err.h>
 #include <openssl/pkcs12.h>
+#include <openssl/x509v3.h>
 #endif
 
 #include "ne_session.h"
@@ -46,8 +49,6 @@
 
 #include "ne_private.h"
 
-#define NEON_USERAGENT "neon/" NEON_VERSION;
-
 #ifdef NEON_SSL
 static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey);
 #endif
@@ -72,7 +73,7 @@ void ne_session_destroy(ne_session *sess)
 
     /* Run the destroy hooks. */
     for (hk = sess->destroy_sess_hooks; hk != NULL; hk = hk->next) {
-	ne_destory_fn fn = (ne_destory_fn)hk->fn;
+	ne_destroy_sess_fn fn = (ne_destroy_sess_fn)hk->fn;
 	fn(hk->userdata);
     }
     
@@ -81,12 +82,14 @@ void ne_session_destroy(ne_session *sess)
     destroy_hooks(sess->post_send_hooks);
     destroy_hooks(sess->destroy_req_hooks);
     destroy_hooks(sess->destroy_sess_hooks);
-    destroy_hooks(sess->accessor_hooks);
+    destroy_hooks(sess->private);
 
     NE_FREE(sess->server.hostname);
     NE_FREE(sess->server.hostport);
+    if (sess->server.address) ne_addr_destroy(sess->server.address);
+    if (sess->proxy.address) ne_addr_destroy(sess->proxy.address);
+    NE_FREE(sess->proxy.hostname);
     NE_FREE(sess->scheme);
-    NE_FREE(sess->proxy.hostport);
     NE_FREE(sess->user_agent);
 
     if (sess->connected) {
@@ -109,36 +112,32 @@ void ne_session_destroy(ne_session *sess)
 
 int ne_version_pre_http11(ne_session *s)
 {
-    return VERSION_PRE11(s);
+    return !s->is_http11;
 }
 
-/* Returns the "hostname[:port]" segment */
-/* FIXME: for SSL. */
-static char *get_hostport(struct host_info *host) 
+/* Stores the "hostname[:port]" segment */
+static void set_hostport(struct host_info *host, unsigned int defaultport)
 {
     size_t len = strlen(host->hostname);
-    char *ret = ne_malloc(len + 10);
-    strcpy(ret, host->hostname);
-    if (host->port != 80) {
-	ne_snprintf(ret + len, 9, ":%d", host->port);
-    }
-    return ret;
+    host->hostport = ne_malloc(len + 10);
+    strcpy(host->hostport, host->hostname);
+    if (host->port != defaultport)
+	ne_snprintf(host->hostport + len, 9, ":%u", host->port);
 }
 
 /* Stores the hostname/port in *info, setting up the "hostport"
  * segment correctly. */
 static void
-set_hostinfo(struct host_info *info, const char *hostname, int port)
+set_hostinfo(struct host_info *info, const char *hostname, unsigned int port)
 {
     NE_FREE(info->hostport);
     NE_FREE(info->hostname);
     info->hostname = ne_strdup(hostname);
     info->port = port;
-    info->hostport = get_hostport(info);
 }
 
 ne_session *ne_session_create(const char *scheme,
-			      const char *hostname, int port)
+			      const char *hostname, unsigned int port)
 {
     ne_session *sess = ne_calloc(sizeof *sess);
 
@@ -146,14 +145,13 @@ ne_session *ne_session_create(const char *scheme,
 	     scheme, hostname, port);
 
     strcpy(sess->error, "Unknown error.");
-    sess->version_major = -1;
-    sess->version_minor = -1;
 
-    /* set the hostname/port */
-    set_hostinfo(&sess->server, hostname, port);
-    
     /* use SSL if scheme is https */
     sess->use_ssl = !strcmp(scheme, "https");
+    
+    /* set the hostname/port */
+    set_hostinfo(&sess->server, hostname, port);
+    set_hostport(&sess->server, sess->use_ssl?443:80);
 
 #ifdef NEON_SSL
     if (sess->use_ssl) {
@@ -170,7 +168,8 @@ ne_session *ne_session_create(const char *scheme,
     return sess;
 }
 
-void ne_session_proxy(ne_session *sess, const char *hostname, int port)
+void ne_session_proxy(ne_session *sess, const char *hostname,
+		      unsigned int port)
 {
     sess->use_proxy = 1;
     set_hostinfo(&sess->proxy, hostname, port);
@@ -181,13 +180,13 @@ void ne_set_error(ne_session *sess, const char *format, ...)
     va_list params;
 
     va_start(params, format);
-    ne_vsnprintf(sess->error, BUFSIZ, format, params);
+    ne_vsnprintf(sess->error, sizeof sess->error, format, params);
     va_end(params);
 }
 
 
 void ne_set_progress(ne_session *sess, 
-		       sock_progress progress, void *userdata)
+		     ne_progress progress, void *userdata)
 {
     sess->progress_cb = progress;
     sess->progress_ud = userdata;
@@ -219,11 +218,13 @@ void ne_set_read_timeout(ne_session *sess, int timeout)
     sess->rdtimeout = timeout;
 }
 
+#define AGENT "neon/" NEON_VERSION " "
+
 void ne_set_useragent(ne_session *sess, const char *token)
 {
-    static const char *fixed = " " NEON_USERAGENT;
-    NE_FREE(sess->user_agent);
-    CONCAT2(sess->user_agent, token, fixed);
+    if (sess->user_agent) free(sess->user_agent);
+    sess->user_agent = malloc(sizeof AGENT + strlen(token));
+    strcat(strcpy(sess->user_agent, AGENT), token);
 }
 
 const char *ne_get_server_hostport(ne_session *sess)
@@ -236,21 +237,40 @@ const char *ne_get_scheme(ne_session *sess)
     return sess->scheme;
 }
 
-const char *ne_get_error(ne_session *sess)
+void ne_fill_server_uri(ne_session *sess, ne_uri *uri)
 {
-    return sess->error;
+    uri->host = ne_strdup(sess->server.hostname);
+    uri->port = sess->server.port;
+    uri->scheme = ne_strdup(sess->scheme);
 }
 
-int ne_close_connection(ne_session *sess)
+/* Cleanse 'str' of non-printable characters.  Duplicated in
+ * ne_utils.c for the duration of neon 0.23.x to prevent ABI
+ * change. */
+static char *cleanse(char *str)
+{
+    char *pnt;
+    for (pnt = str; *pnt; pnt++)
+        if (iscntrl(*pnt) || !isprint(*pnt)) *pnt = ' ';
+    return str;
+}
+
+const char *ne_get_error(ne_session *sess)
+{
+    return cleanse(sess->error);
+}
+
+void ne_close_connection(ne_session *sess)
 {
-    NE_DEBUG(NE_DBG_SOCKET, "Closing connection.\n");
     if (sess->connected > 0) {
-	sock_close(sess->socket);
+	NE_DEBUG(NE_DBG_SOCKET, "Closing connection.\n");
+	ne_sock_close(sess->socket);
 	sess->socket = NULL;
+	NE_DEBUG(NE_DBG_SOCKET, "Connection closed.\n");
+    } else {
+	NE_DEBUG(NE_DBG_SOCKET, "(Not closing closed connection!).\n");
     }
     sess->connected = 0;
-    NE_DEBUG(NE_DBG_SOCKET, "Connection closed.\n");
-    return 0;
 }
 
 void ne_ssl_set_verify(ne_session *sess, ne_ssl_verify_fn fn, void *userdata)
@@ -266,6 +286,32 @@ void ne_ssl_provide_ccert(ne_session *sess,
     sess->ssl_provide_ud = userdata;
 }
 
+void ne_ssl_keypw_prompt(ne_session *sess, ne_ssl_keypw_fn fn, void *ud)
+{
+    sess->ssl_keypw_fn = fn;
+    sess->ssl_keypw_ud = ud;
+}
+
+char *ne_ssl_readable_dname(const ne_ssl_dname *dn)
+{
+    int flag = 0;
+    ne_buffer *buf = ne_buffer_create();
+
+#define DO(f) do { \
+if (f) { ne_buffer_concat(buf, flag?", ":"", f, NULL); flag = 1; } \
+} while (0)
+
+    DO(dn->organizationalUnit);
+    DO(dn->organization);
+    DO(dn->locality);
+    DO(dn->state);
+    DO(dn->country);
+
+#undef DO
+
+    return ne_buffer_finish(buf);
+}
+
 #ifdef NEON_SSL
 
 SSL_CTX *ne_ssl_get_context(ne_session *sess)
@@ -293,8 +339,8 @@ static void verify_err(ne_session *sess, int failures)
 
     for (n = 0; reasons[n].bit; n++) {
 	if (failures & reasons[n].bit) {
-	    if (flag) strncat(sess->error, ", ", BUFSIZ);
-	    strncat(sess->error, _(reasons[n].str), BUFSIZ);
+	    if (flag) strncat(sess->error, ", ", sizeof sess->error);
+	    strncat(sess->error, _(reasons[n].str), sizeof sess->error);
 	    flag = 1;
 	}
     }
@@ -324,33 +370,35 @@ static const char *getx509field(X509_NAME *name, int nid,
     }
 }
 
-/* Return malloc-allocated string representatino of given ASN.1 time
+/* Return malloc-allocated string representation of given ASN.1 time
  * structure.  TODO: would be better to parse out the raw ASN.1 string
  * and give that to the application in some form which is localisable
  * e.g. time_t.  */
 static char *asn1time_to_string(ASN1_TIME *tm)
 {
-  char buf[64];
-  BIO *bio;
+    char buf[64];
+    BIO *bio;
+    
+    strncpy(buf, _("[invalid date]"), sizeof buf);
+    
+    bio = BIO_new(BIO_s_mem());
+    if (bio) {
+	if (ASN1_TIME_print(bio, tm))
+	    BIO_read(bio, buf, sizeof buf);
+	BIO_free(bio);
+    }
   
-  strncpy(buf, _("[invalid date]"), sizeof(buf));
-
-  bio = BIO_new(BIO_s_mem());
-  if (bio) {
-      if (ASN1_TIME_print(bio, tm))
-	  BIO_read(bio, buf, sizeof(buf));
-      BIO_free(bio);
-  }
-  
-  return ne_strdup(buf);
+    return ne_strdup(buf);
 }
 
-/* Return non-zero if hostname from certificate (cn) doesn't match
- * hostname used for session (hostname). TODO: could do more advanced
- * wildcard matching using fnmatch() here, if fnmatch is present. */
+/* Return non-zero if hostname from certificate (cn) matches hostname
+ * used for session (hostname). TODO: could do more advanced wildcard
+ * matching using fnmatch() here, if fnmatch is present. */
 static int match_hostname(char *cn, const char *hostname)
 {
-    const char *dot = strchr(hostname, '.');
+    const char *dot;
+    NE_DEBUG(NE_DBG_SSL, "Match %s on %s...\n", cn, hostname);
+    dot = strchr(hostname, '.');
     if (dot == NULL) {
 	char *pnt = strchr(cn, '.');
 	/* hostname is not fully-qualified; unqualify the cn. */
@@ -362,7 +410,7 @@ static int match_hostname(char *cn, const char *hostname)
 	hostname = dot + 1;
 	cn += 2;
     }
-    return strcmp(cn, hostname);
+    return !strcasecmp(cn, hostname);
 }
 
 /* Fills in the friendly DN structure 'dn' from given X509 name 'xn',
@@ -378,6 +426,68 @@ static void make_dname(ne_ssl_dname *dn, X509_NAME *xn, ne_buffer *dump)
     dn->commonName = getx509field(xn, NID_commonName, dump);
 }
 
+/* Check certificate identity.  Returns zero if identity matches; 1 if
+ * identity does not match, or <0 on error (session error string
+ * set). */
+static int check_identity(ne_session *sess, X509 *cert)
+{
+    STACK_OF(GENERAL_NAME) *names;
+    int match = 0, found = 0;
+    
+    names = X509_get_ext_d2i(cert, NID_subject_alt_name, NULL, NULL);
+    if (names) {
+	/* Got a subject alt. name extension. */
+	int n;
+
+	for (n = 0; n < sk_GENERAL_NAME_num(names) && !match; n++) {
+	    GENERAL_NAME *nm = sk_GENERAL_NAME_value(names, n);
+	    
+	    /* only care about this if it is a DNS name. */
+	    if (nm->type == GEN_DNS) {
+		/* TODO: convert this string to UTF8 or something. */
+		char *name = ne_strndup(nm->d.ia5->data, nm->d.ia5->length);
+		match = match_hostname(name, sess->server.hostname);
+		free(name);
+		found = 1;
+	    }
+	}
+	sk_GENERAL_NAME_free(names);
+    }
+    
+    /* Check against the commonName if no DNS alt. names were found,
+     * as per RFC2818. */
+    if (!found) {
+	X509_NAME *subj = X509_get_subject_name(cert);
+	X509_NAME_ENTRY *entry;
+	ASN1_STRING *str;
+	int idx = -1, lastidx;
+	char *name;
+
+	/* find the most specific commonName attribute. */
+	do {
+	    lastidx = idx;
+	    idx = X509_NAME_get_index_by_NID(subj, NID_commonName, lastidx);
+	} while (idx >= 0);
+	
+	if (lastidx < 0) {
+	    ne_set_error(sess, _("Server certificate was missing "
+				 "commonName attribute in subject name"));
+	    return -1;
+	}
+
+	/* extract the string from the entry */
+	entry = X509_NAME_get_entry(subj, lastidx);
+	str = X509_NAME_ENTRY_get_data(entry);
+
+	name = ne_strndup(str->data, str->length);
+	match = match_hostname(name, sess->server.hostname);
+	free(name);
+    }
+
+    NE_DEBUG(NE_DBG_SSL, "Identity match: %s\n", match ? "good" : "bad");
+    return match ? 0 : 1;
+}
+
 /* Verifies an SSL server certificate. */
 static int check_certificate(ne_session *sess, SSL *ssl, X509 *cert)
 {
@@ -385,39 +495,39 @@ static int check_certificate(ne_session *sess, SSL *ssl, X509 *cert)
     X509_NAME *issuer = X509_get_issuer_name(cert);
     ASN1_TIME *notBefore = X509_get_notBefore(cert);
     ASN1_TIME *notAfter = X509_get_notAfter(cert);
-    char buf[ATTBUFSIZ];
     int ret, failures = 0;
     long result;
 
     /* check expiry dates */
-    if (X509_cmp_current_time(notBefore) >= 0) {
+    if (X509_cmp_current_time(notBefore) >= 0)
 	failures |= NE_SSL_NOTYETVALID;
-    }
-    else if (X509_cmp_current_time(notAfter) <= 0) {
+    else if (X509_cmp_current_time(notAfter) <= 0)
 	failures |= NE_SSL_EXPIRED;
-    } 
 
-    /* retrieve the commonName, compare with the server hostname. */
-    ret = X509_NAME_get_text_by_NID(subj, NID_commonName, buf, ATTBUFSIZ);
-    if (ret < 1) {
-	ne_set_error(sess, 
-		     _("Server certificate was missing commonName attribute"));
-	return NE_ERROR;
-    }
-    
-    if (match_hostname(buf, sess->server.hostname)) {
-	failures |= NE_SSL_CNMISMATCH;
-    }
+    /* Check certificate was issued to this server. */
+    ret = check_identity(sess, cert);
+    if (ret < 0) return NE_ERROR;
+    else if (ret > 0) failures |= NE_SSL_CNMISMATCH;
 
-    /* get the result of the cert verication out of OpenSSL */
+    /* get the result of the cert verification out of OpenSSL */
     result = SSL_get_verify_result(ssl);
 
-    NE_DEBUG(NE_DBG_HTTP, "Verify result: %ld = %s\n", result,
+    NE_DEBUG(NE_DBG_SSL, "Verify result: %ld = %s\n", result,
 	     X509_verify_cert_error_string(result));
 
 #if NE_DEBUGGING
-    if (ne_debug_mask & NE_DBG_HTTP)
-	X509_print_fp(ne_debug_stream, cert);
+    if (ne_debug_mask & NE_DBG_SSL) {
+	STACK_OF(X509) *chain = SSL_get_peer_cert_chain(ssl);
+	int n, top = chain?sk_X509_num(chain):0;
+
+	NE_DEBUG(NE_DBG_SSL, "Peer chain depth is %d\n", top);
+	
+	for (n = 0; n < top; n++) {
+	    X509 *c = sk_X509_value(chain, n);
+	    NE_DEBUG(NE_DBG_SSL, "Cert #%d:\n", n);
+	    X509_print_fp(ne_debug_stream, c);
+	}
+    }
 #endif
 
     switch (result) {
@@ -450,7 +560,7 @@ static int check_certificate(ne_session *sess, SSL *ssl, X509 *cert)
     if (sess->ssl_verify_fn && failures) {
 	ne_ssl_certificate c;
 	ne_ssl_dname sdn = {0}, idn = {0};
-	ne_buffer *dump = ne_buffer_create_sized(ATTBUFSIZ * 2);
+	ne_buffer *dump = ne_buffer_ncreate(ATTBUFSIZ * 2);
 	char *from, *until;
 	
 	/* Do the gymnatics to retrieve attributes out of the
@@ -498,26 +608,35 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
 
     if (!sess->client_key && sess->ssl_provide_fn) {
 	ne_ssl_dname dn;
-	ne_buffer *buf = ne_buffer_create_sized(2048);
+	ne_buffer *buf;
 	X509 *peer = SSL_get_peer_certificate(ssl);
-	X509_NAME *subject = peer?X509_get_subject_name(peer):NULL;
-	
-	if (subject) {
-	    make_dname(&dn, subject, buf);
+	X509_NAME *subject;	
+
+	if (!peer) {
+	    NE_DEBUG(NE_DBG_SSL, 
+		     "Peer subject unspecified; cannot provide c.cert\n");
+	    return 0;
 	}
 
-	NE_DEBUG(NE_DBG_HTTP, "Calling client certificate provider...\n");
+	subject = X509_get_subject_name(peer);
+
+	buf = ne_buffer_ncreate(2048);
+	make_dname(&dn, subject, buf);
+
+	NE_DEBUG(NE_DBG_SSL, "Calling client certificate provider...\n");
 	sess->ssl_provide_fn(sess->ssl_provide_ud, sess, &dn);
 	ne_buffer_destroy(buf);			     
     }
 
     if (sess->client_key && sess->client_cert) {
-	NE_DEBUG(NE_DBG_HTTP, "Supplying client certificate.");
+	NE_DEBUG(NE_DBG_SSL, "Supplying client certificate.\n");
+	sess->client_cert->references++;
+	sess->client_key->references++;
 	*cert = sess->client_cert;
 	*pkey = sess->client_key;
 	return 1;
     } else {
-	NE_DEBUG(NE_DBG_HTTP, "No client certificate supplied.\n");
+	NE_DEBUG(NE_DBG_SSL, "No client certificate supplied.\n");
 	return 0;
     }
 }
@@ -525,22 +644,21 @@ static int provide_client_cert(SSL *ssl, X509 **cert, EVP_PKEY **pkey)
 /* For internal use only. */
 int ne_negotiate_ssl(ne_request *req)
 {
-    ne_session *sess = req->session;
+    ne_session *sess = ne_get_session(req);
     SSL *ssl;
     X509 *cert;
 
-    NE_DEBUG(NE_DBG_HTTP, "Doing SSL negotiation.\n");
+    NE_DEBUG(NE_DBG_SSL, "Doing SSL negotiation.\n");
 
-    if (sock_enable_ssl_os(sess->socket, sess->ssl_context, 
+    if (ne_sock_use_ssl_os(sess->socket, sess->ssl_context, 
 			   sess->ssl_sess, &ssl, sess)) {
-	NE_DEBUG(NE_DBG_HTTP, "sock_eso failed: %s\n", 
-		 sock_get_error(sess->socket));
 	if (sess->ssl_sess) {
 	    /* remove cached session. */
 	    SSL_SESSION_free(sess->ssl_sess);
 	    sess->ssl_sess = NULL;
 	}
-	ne_set_error(sess, _("SSL negotiation failed"));
+	ne_set_error(sess, _("SSL negotiation failed: %s"),
+		     ne_sock_error(sess->socket));
 	return NE_ERROR;
     }	
     
@@ -551,22 +669,24 @@ int ne_negotiate_ssl(ne_request *req)
     }
 
     if (sess->server_cert) {
-	if (X509_cmp(cert, sess->server_cert)) {
+	int cmp = X509_cmp(cert, sess->server_cert);
+	X509_free(cert);
+	if (cmp) {
 	    /* This could be a MITM attack: fail the request. */
-	    ne_set_error(sess, _("Server certificated changed: connection intercepted?"));
+	    ne_set_error(sess, _("Server certificate changed: "
+				 "connection intercepted?"));
 	    X509_free(sess->server_cert);
 	    sess->server_cert = NULL;
 	    return NE_ERROR;
 	} 
-	/* server_cert is the same as cert, so abandon new copy. */
-	X509_free(cert);
 	/* certificate has already passed verification: no need to
 	 * verify it again. */
     } else {
 	/* new connection: verify the cert. */
 	if (check_certificate(sess, ssl, cert)) {
-	    NE_DEBUG(NE_DBG_HTTP, "SSL certificate checks failed: %s\n",
+	    NE_DEBUG(NE_DBG_SSL, "SSL certificate checks failed: %s\n",
 		     sess->error);
+	    X509_free(cert);
 	    return NE_ERROR;
 	}
 	/* cache the cert. */
@@ -608,16 +728,6 @@ static int privkey_prompt(char *buf, int len, int rwflag, void *userdata)
     return strlen(buf);
 }
 
-void ne_ssl_keypw_prompt(ne_session *sess, ne_ssl_keypw_fn fn, void *ud)
-{
-    /* install our wrapper callback. */
-    SSL_CTX_set_default_passwd_cb(sess->ssl_context, privkey_prompt);
-    SSL_CTX_set_default_passwd_cb_userdata(sess->ssl_context, sess);
-    /* and remember the application callback. */
-    sess->ssl_keypw_fn = fn;
-    sess->ssl_keypw_ud = ud;
-}
-
 int ne_ssl_load_pkcs12(ne_session *sess, const char *fn)
 {
     /* you are lost in a maze of twisty crypto algorithms... */
@@ -628,10 +738,10 @@ int ne_ssl_load_pkcs12(ne_session *sess, const char *fn)
 
     fp = fopen(fn, "r");
     if (fp == NULL) {
-	ne_set_error(sess, _("Could not open file `%s': %s"), fn,
-		     strerror(errno));
+	ne_strerror(errno, buf, sizeof buf);
+	ne_set_error(sess, _("Could not open file `%s': %s"), fn, buf);
 	return -1;
-    }		     
+    }
 
     p12 = d2i_PKCS12_fp(fp, NULL);
 
@@ -647,7 +757,7 @@ int ne_ssl_load_pkcs12(ne_session *sess, const char *fn)
     sess->client_cert = NULL;
 
     if (sess->ssl_keypw_fn) {
-	if (sess->ssl_keypw_fn(sess->ssl_keypw_ud, buf, BUFSIZ) == 0)
+	if (sess->ssl_keypw_fn(sess->ssl_keypw_ud, buf, sizeof buf) == 0)
 	    password = buf;
     }
 	
@@ -668,20 +778,22 @@ int ne_ssl_load_pkcs12(ne_session *sess, const char *fn)
 int ne_ssl_load_pem(ne_session *sess, const char *cert, const char *key)
 {
     FILE *fp;
+    char err[200];
 
     sess->client_key = NULL;
     sess->client_cert = NULL;
 
     fp = fopen(cert, "r");
     if (fp == NULL) {
-	ne_set_error(sess, _("Could not open file `%s': %s"), cert,
-		     strerror(errno));
+	ne_strerror(errno, err, sizeof err);
+	ne_set_error(sess, _("Could not open file `%s': %s"), cert, err);
 	return -1;
     }
 
-    sess->client_cert = PEM_read_X509(fp, NULL, NULL, NULL);
+    sess->client_cert = PEM_read_X509(fp, NULL, privkey_prompt, sess);
     if (sess->client_cert == NULL) {
-	ne_set_error(sess, _("Could not read certificate"));
+	ne_set_error(sess, _("Could not read certificate: %s"),
+		     ERR_reason_error_string(ERR_get_error()));
 	fclose(fp);
 	return -1;
     }
@@ -690,17 +802,18 @@ int ne_ssl_load_pem(ne_session *sess, const char *cert, const char *key)
 	fclose(fp);
 	fp = fopen(key, "r");
 	if (fp == NULL) {
-	    ne_set_error(sess, 
-			 _("Could not open private key file `%s': %s"),
-			 key, strerror(errno));
+	    ne_strerror(errno, err, sizeof err);
+	    ne_set_error(sess, _("Could not open private key file `%s': %s"),
+			 key, err);
 	    return -1;
 	}
     }
 
-    sess->client_key = PEM_read_PrivateKey(fp, NULL, NULL, NULL);
+    sess->client_key = PEM_read_PrivateKey(fp, NULL, privkey_prompt, sess);
+    fclose(fp);
     if (sess->client_key == NULL) {
-	ne_set_error(sess, 
-		     _("Could not parse private key (incorrect password?)"));
+	ne_set_error(sess, _("Could not parse private key: %s"),
+		     ERR_reason_error_string(ERR_get_error()));
 	return -1;
     }
     
@@ -718,12 +831,11 @@ X509 *ne_ssl_server_cert(ne_session *sess)
 
 /* Stubs to make the library have the same ABI whether or not SSL
  * support is enabled. */
-int ne_negotiate_ssl(ne_request *req) { STUB(req->session); }
+int ne_negotiate_ssl(ne_request *req) { STUB(ne_get_session(req)); }
 int ne_ssl_load_ca(ne_session *sess, const char *file) { STUB(sess); }
 int ne_ssl_load_default_ca(ne_session *sess) { STUB(sess); }
 int ne_ssl_load_pkcs12(ne_session *sess, const char *fn) { STUB(sess); }
 int ne_ssl_load_pem(ne_session *sess, const char *cert, const char *key) { STUB(sess); }
-void ne_ssl_keypw_prompt(ne_session *sess, ne_ssl_keypw_fn fn, void *ud) {}
 
 #endif
 
