@@ -36,48 +36,49 @@
 #include "ne_alloc.h"
 #include "ne_string.h"
 
-char *ne_token(char **str, char separator, const char *quotes)
+char *ne_token(char **str, char separator)
+{
+    char *ret = *str, *pnt = strchr(*str, separator);
+
+    if (pnt) {
+	*pnt = '\0';
+	*str = pnt + 1;
+    } else {
+	/* no separator found: return end of string. */
+	*str = NULL;
+    }
+    
+    return ret;
+}
+
+char *ne_qtoken(char **str, char separator, const char *quotes)
 {
     char *pnt, *ret = NULL;
 
-    if (quotes) {
-
-	for (pnt = *str; *pnt != '\0'; pnt++) {
-	    char *quot = strchr(quotes, *pnt);
+    for (pnt = *str; *pnt != '\0'; pnt++) {
+	char *quot = strchr(quotes, *pnt);
+	
+	if (quot) {
+	    char *qclose = strchr(pnt+1, *quot);
 	    
-	    if (quot) {
-		char *qclose = strchr(pnt+1, *quot);
-		
-		if (!qclose) {
-		    /* no closing quote: invalid string. */
-		    return NULL;
-		}
-		
-		pnt = qclose;
-	    } else if (*pnt == separator) {
-		/* found end of token. */
-		*pnt = '\0';
-		ret = *str;
-		*str = pnt + 1;
-		break;
+	    if (!qclose) {
+		/* no closing quote: invalid string. */
+		return NULL;
 	    }
-	}
-
-    } else {
-	pnt = strchr(*str, separator);
-	if (pnt != NULL) {
+	    
+	    pnt = qclose;
+	} else if (*pnt == separator) {
+	    /* found end of token. */
 	    *pnt = '\0';
 	    ret = *str;
 	    *str = pnt + 1;
+	    return ret;
 	}
     }
 
-    if (ret == NULL) {
-	/* no separator found: return end of string. */
-	ret = *str;
-	*str = NULL;
-    }
-
+    /* no separator found: return end of string. */
+    ret = *str;
+    *str = NULL;
     return ret;
 }
 
@@ -266,23 +267,6 @@ void pair_string_free(char **pairs)
     free(pairs);
 }
 
-char *shave_string(const char *str, const char ch) 
-{
-    size_t len = strlen(str);
-    char *ret;
-    if (str[len-1] == ch) {
-	len--;
-    }
-    if (str[0] == ch) {
-	len--;
-	str++;
-    }
-    ret = ne_malloc(len + 1);
-    memcpy(ret, str, len);
-    ret[len] = '\0';
-    return ret;
-}
-
 void ne_buffer_clear(ne_buffer *buf) 
 {
     memset(buf->data, 0, buf->length);
@@ -302,40 +286,66 @@ void ne_buffer_grow(ne_buffer *buf, size_t newsize)
     }
 }
 
-void ne_buffer_concat(ne_buffer *buf, ...) 
+static size_t count_concat(va_list *ap)
+{
+    size_t total = 0;
+    char *next;
+
+    while ((next = va_arg(*ap, char *)) != NULL)
+	total += strlen(next);
+
+    return total;
+}
+
+static void do_concat(char *str, va_list *ap) 
+{
+    char *next;
+
+    while ((next = va_arg(*ap, char *)) != NULL) {
+	size_t len = strlen(next);
+	memcpy(str, next, len);
+	str += len;
+    }
+}
+
+void ne_buffer_concat(ne_buffer *buf, ...)
 {
     va_list ap;
-    char *next, *pnt;
-    size_t totallen = buf->used; 
+    ssize_t total;
 
-    /* Find out how much space we need for all the args */
     va_start(ap, buf);
-    do {
-	next = va_arg(ap, char *);
-	if (next != NULL) {
-	    totallen += strlen(next);
-	}
-    } while (next != NULL);
-    va_end(ap);
-    
+    total = buf->used + count_concat(&ap);
+    va_end(ap);    
+
     /* Grow the buffer */
-    ne_buffer_grow(buf, totallen);
+    ne_buffer_grow(buf, total);
     
-    /* Now append the arguments to the buffer */
-    va_start(ap, buf);
-    pnt = buf->data + buf->used - 1;
-    do {
-	next = va_arg(ap, char *);
-	if (next != NULL) {
-	    size_t len = strlen(next);
-	    memcpy(pnt, next, len);
-	    pnt += len;
-	}
-    } while (next != NULL);
+    va_start(ap, buf);    
+    do_concat(buf->data + buf->used - 1, &ap);
+    va_end(ap);    
+
+    buf->used = total;
+    buf->data[total - 1] = '\0';
+}
+
+char *ne_concat(const char *str, ...)
+{
+    va_list ap;
+    size_t total, slen = strlen(str);
+    char *ret;
+
+    va_start(ap, str);
+    total = slen + count_concat(&ap);
     va_end(ap);
-    
-    buf->used = totallen;
-    buf->data[totallen - 1] = '\0';
+
+    ret = memcpy(ne_malloc(total + 1), str, slen);
+
+    va_start(ap, str);
+    do_concat(ret + slen, &ap);
+    va_end(ap);
+
+    ret[total] = '\0';
+    return ret;    
 }
 
 /* Append zero-terminated string... returns 0 on success or -1 on
@@ -355,12 +365,12 @@ void ne_buffer_append(ne_buffer *buf, const char *data, size_t len)
 
 ne_buffer *ne_buffer_create(void) 
 {
-    return ne_buffer_create_sized(512);
+    return ne_buffer_ncreate(512);
 }
 
-ne_buffer *ne_buffer_create_sized(size_t s) 
+ne_buffer *ne_buffer_ncreate(size_t s) 
 {
-    ne_buffer *buf = ne_malloc(sizeof(struct ne_buffer_s));
+    ne_buffer *buf = ne_malloc(sizeof(*buf));
     buf->data = ne_malloc(s);
     buf->data[0] = '\0';
     buf->length = s;
@@ -370,9 +380,7 @@ ne_buffer *ne_buffer_create_sized(size_t s)
 
 void ne_buffer_destroy(ne_buffer *buf) 
 {
-    if (buf->data) {
-	free(buf->data);
-    }
+    free(buf->data);
     free(buf);
 }
 
@@ -388,51 +396,66 @@ void ne_buffer_altered(ne_buffer *buf)
     buf->used = strlen(buf->data) + 1;
 }
 
-char *ne_utf8_encode(const char *str)
+static const char *b64_alphabet =  
+    "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
+    "abcdefghijklmnopqrstuvwxyz"
+    "0123456789+/=";
+    
+char *ne_base64(const char *text, size_t inlen)
 {
-    char *buffer = ne_malloc(strlen(str) * 2 + 1);
-    int in, len = strlen(str);
-    char *out;
+    /* The tricky thing about this is doing the padding at the end,
+     * doing the bit manipulation requires a bit of concentration only */
+    char *buffer, *point;
+    size_t outlen;
+    
+    /* Use 'buffer' to store the output. Work out how big it should be...
+     * This must be a multiple of 4 bytes */
 
-    for (in = 0, out = buffer; in < len; in++, out++) {
-	if ((unsigned char)str[in] <= 0x7D) {
-	    *out = str[in];
-	} else {
-	    *out++ = 0xC0 | ((str[in] & 0xFC) >> 6);
-	    *out = str[in] & 0xBF;
-	}
+    outlen = (inlen*4)/3;
+    if ((inlen % 3) > 0) /* got to pad */
+	outlen += 4 - (inlen % 3);
+    
+    buffer = ne_malloc(outlen + 1); /* +1 for the \0 */
+    
+    /* now do the main stage of conversion, 3 bytes at a time,
+     * leave the trailing bytes (if there are any) for later */
+
+    for (point=buffer; inlen>=3; inlen-=3, text+=3) {
+	*(point++) = b64_alphabet[ (*text)>>2 ]; 
+	*(point++) = b64_alphabet[ ((*text)<<4 & 0x30) | (*(text+1))>>4 ]; 
+	*(point++) = b64_alphabet[ ((*(text+1))<<2 & 0x3c) | (*(text+2))>>6 ];
+	*(point++) = b64_alphabet[ (*(text+2)) & 0x3f ];
     }
 
-    /* nul-terminate */
-    *out = '\0';
+    /* Now deal with the trailing bytes */
+    if (inlen > 0) {
+	/* We always have one trailing byte */
+	*(point++) = b64_alphabet[ (*text)>>2 ];
+	*(point++) = b64_alphabet[ (((*text)<<4 & 0x30) |
+				     (inlen==2?(*(text+1))>>4:0)) ]; 
+	*(point++) = (inlen==1?'=':b64_alphabet[ (*(text+1))<<2 & 0x3c ]);
+	*(point++) = '=';
+    }
+
+    /* Null-terminate */
+    *point = '\0';
+
     return buffer;
 }
 
-/* Single byte range 0x00 -> 0x7F */
-#define SINGLEBYTE_UTF8(ch) (((unsigned char) (ch)) < 0x80)
-
-char *ne_utf8_decode(const char *str)
+char *ne_strerror(int errnum, char *buf, size_t buflen)
 {
-    int n, m, len = strlen(str);
-    char *dest = ne_malloc(len + 1);
-    
-    for (n = 0, m = 0; n < len; n++, m++) {
-	if (SINGLEBYTE_UTF8(str[n])) {
-	    dest[m] = str[n];
-	} else {
-	    /* We only deal with 8-bit data, which will be encoded as
-	     * two bytes of UTF-8 */
-	    if ((len - n < 2) || (str[n] & 0xFC) != 0xC0) {
-		free(dest);
-		return NULL;
-	    } else {
-		/* see hip_xml.c for comments. */
-		dest[m] = ((str[n] & 0x03) << 6) | (str[n+1] & 0x3F);
-		n++;
-	    }
-	}
-    }
-    
-    dest[m] = '\0';
-    return dest;
+#ifdef HAVE_STRERROR_R
+#ifdef STRERROR_R_CHAR_P
+    /* glibc-style strerror_r which may-or-may-not use provided buffer. */
+    char *ret = strerror_r(errnum, buf, buflen);
+    if (ret != buf)
+	ne_strnzcpy(buf, ret, buflen);
+#else /* POSIX-style strerror_r: */
+    strerror_r(errnum, buf, buflen);
+#endif
+#else /* no strerror_r: */
+    ne_strnzcpy(buf, strerror(errnum), buflen);
+#endif
+    return buf;
 }
