@@ -5,10 +5,15 @@ rm -f child.log
 
 # for shared builds.
 LD_LIBRARY_PATH=../src/.libs:$LD_LIBRARY_PATH
-export LD_LIBRARY_PATH
+
+# enable an safety-checking malloc in glibc which will abort() if
+# heap corruption is detected.
+MALLOC_CHECK_=2
+
+export LD_LIBRARY_PATH MALLOC_CHECK_
 
 for f in $*; do
-    if ./$f ${SRCDIR}; then
+    if ${HARNESS} ./$f ${SRCDIR}; then
 	:
     else
 	echo FAILURE
