@@ -52,9 +52,9 @@
 #define SERVER_CERT "server.cert"
 #define CA_CERT "ca/cert.pem"
 
-SSL_CTX *server_ctx;
+static SSL_CTX *server_ctx = NULL;
 
-char *srcdir = ".";
+static char *srcdir = ".";
 
 static int s_strwrite(SSL *s, const char *buf)
 {
@@ -68,10 +68,10 @@ static int s_strwrite(SSL *s, const char *buf)
 
 /* Do an SSL response over socket given context; returning ssl session
  * structure in *sess if sess is non-NULL. */
-static int do_ssl_response(nsocket *sock, SSL_CTX *ctx,
-			   SSL_SESSION **sess)
+static int do_ssl_response(ne_socket *sock, SSL_CTX *ctx, SSL_SESSION **sess,
+			   const char *resp, int unclean)
 {
-    int fd = sock_get_fd(sock);
+    int fd = ne_sock_fd(sock), ret;
     /* we don't want OpenSSL to close this socket for us. */
     BIO *bio = BIO_new_socket(fd, BIO_NOCLOSE);
     char buf[BUFSIZ];
@@ -84,30 +84,34 @@ static int do_ssl_response(nsocket *sock, SSL_CTX *ctx,
     ONV(SSL_accept(ssl) != 1,
 	("SSL_accept failed: %s", ERROR_SSL_STRING));
 
-    ONV(SSL_read(ssl, buf, BUFSIZ) <= 0,
-	("SSL_read failed: %s", ERROR_SSL_STRING));
+    ret = SSL_read(ssl, buf, BUFSIZ);
+    if (ret == 0)
+	return 0; /* connection closed by parent; give up. */
+    ONV(ret < 0, ("SSL_read failed (%d): %s", ret, ERROR_SSL_STRING));
 
-    CALL(s_strwrite(ssl, "HTTP/1.0 200 OK\r\n"
-		    "Content-Length: 0\r\n"
-		    "Connection: close\r\n\r\n"));
+    CALL(s_strwrite(ssl, resp));
 
     /* copy out the session if requested. */
     if (sess) {
 	*sess = SSL_get1_session(ssl);
     }	
     
-    /* Erk, shutdown is messy! See Eric Rescorla's article:
-     * http://www.linuxjournal.com/article.php?sid=4822 ; we'll just
-     * hide our heads in the sand here. */
-    SSL_shutdown(ssl);
-    SSL_free(ssl);
+    if (!unclean) {
+	/* Erk, shutdown is messy! See Eric Rescorla's article:
+	 * http://www.linuxjournal.com/article.php?sid=4822 ; we'll just
+	 * hide our heads in the sand here. */
+	SSL_shutdown(ssl);
+	SSL_free(ssl);
+    }
 
     return 0;
 }
 
+#define DEF_RESP "HTTP/1.0 200 OK\r\nContent-Length: 0\r\n\r\n"
+
 /* Standard server callback to send an HTTP response; SSL negotiated
  * using certificate passed as userdata. */
-static int serve_ssl(nsocket *sock, void *ud)
+static int serve_ssl(ne_socket *sock, void *ud)
 {
     const char *cert = ud;
 
@@ -116,13 +120,26 @@ static int serve_ssl(nsocket *sock, void *ud)
     ONN("failed to load certificate",
 	!SSL_CTX_use_certificate_file(server_ctx, cert, SSL_FILETYPE_PEM));
 
-    CALL(do_ssl_response(sock, server_ctx, NULL));
+    CALL(do_ssl_response(sock, server_ctx, NULL, DEF_RESP, 0));
 
     return OK;
 }
 
+static int serve_response_unclean(ne_socket *sock, void *ud)
+{
+    const char *resp = ud;
+
+    ONN("failed to load certificate",
+	!SSL_CTX_use_certificate_file(server_ctx, 
+				      SERVER_CERT, SSL_FILETYPE_PEM));
+
+    CALL(do_ssl_response(sock, server_ctx, NULL, resp, 1));
+
+    return OK;
+}    
+
 /* Server function which requires use of a client cert. */
-static int serve_ccert(nsocket *sock, void *ud)
+static int serve_ccert(ne_socket *sock, void *ud)
 {
     const char *cert = ud;
 
@@ -143,22 +160,14 @@ static int serve_ccert(nsocket *sock, void *ud)
     SSL_CTX_set_client_CA_list(server_ctx, 
 			       SSL_load_client_CA_file(CA_CERT));
 
-    CALL(do_ssl_response(sock, server_ctx, NULL));
+    CALL(do_ssl_response(sock, server_ctx, NULL, DEF_RESP, 0));
 
     return OK;
 }
 
-static int fail_serve(nsocket *sock, void *ud)
+static int fail_serve(ne_socket *sock, void *ud)
 {
-    const char *cert = ud;
-
-    NE_DEBUG(NE_DBG_HTTP, "using server cert %s\n", cert);
-
-    ONN("failed to load certificate",
-	!SSL_CTX_use_certificate_file(server_ctx, cert, SSL_FILETYPE_PEM));
-
-    do_ssl_response(sock, server_ctx, NULL);
-
+    serve_ssl(sock, ud);
     return OK;
 }
 
@@ -195,17 +204,25 @@ static int init(void)
     /* take srcdir as argv[1]. */
     if (test_argc > 1) {
 	srcdir = test_argv[1];
-	CONCAT2(server_key, srcdir, "/server.key");
+	server_key = ne_concat(srcdir, "/server.key", NULL);
     } else {
 	server_key = "server.key";
     }
     
-    ONN("sock_init failed.\n", sock_init());
+    if (ne_sock_init()) {
+	t_context("could not initialize socket/SSL library.");
+	return FAILHARD;
+    }
+
     server_ctx = SSL_CTX_new(SSLv23_server_method());
-    ONN("SSL_CTX_new failed", server_ctx == NULL);
-    ONN("failed to load private key",
-	!SSL_CTX_use_PrivateKey_file(server_ctx, server_key, 
-				     SSL_FILETYPE_PEM));
+    if (server_ctx == NULL) {
+	t_context("could not create SSL_CTX: %s", ERROR_SSL_STRING);
+	return FAILHARD;
+    } else if (!SSL_CTX_use_PrivateKey_file(server_ctx, server_key, 
+					    SSL_FILETYPE_PEM)) {
+	t_context("failed to load private key: %s", ERROR_SSL_STRING);
+	return FAILHARD;
+    }
     return OK;
 }
 
@@ -261,7 +278,6 @@ static int load_pkcs12_ccert(void)
     return OK;
 }
 
-#if 0
 static int fail_load_ccerts(void)
 {
     ne_session *sess = DEFSESS;
@@ -275,19 +291,86 @@ static int fail_load_ccerts(void)
     ne_session_destroy(sess);
     return OK;
 }
-#endif
+
+/* Test that 'cert', which is signed by CA_CERT, is accepted
+ * unconditionaly. */
+static int accept_signed_cert(char *cert)
+{
+    ne_session *sess = DEFSESS;
+    /* no verify callback needed. */
+    CALL(any_ssl_request(sess, serve_ssl, cert, CA_CERT, NULL, NULL));
+    ne_session_destroy(sess);
+    return OK;
+}
 
 static int simple(void)
 {
-    ne_session *sess = DEFSESS;
+    return accept_signed_cert(SERVER_CERT);
+}
 
-    /* CA cert is trusted, so no verify callback should be needed. */
-    CALL(any_ssl_request(sess, serve_ssl, SERVER_CERT, CA_CERT, NULL, NULL));
-    ne_session_destroy(sess);
+/* Serves using HTTP/1.0 get-till-EOF semantics. */
+static int serve_eof(ne_socket *sock, void *ud)
+{
+    const char *cert = ud;
+
+    NE_DEBUG(NE_DBG_HTTP, "using server cert %s\n", cert);
+
+    ONN("failed to load certificate",
+	!SSL_CTX_use_certificate_file(server_ctx, cert, SSL_FILETYPE_PEM));
+
+    CALL(do_ssl_response(sock, server_ctx, NULL,
+			 "HTTP/1.0 200 OK\r\n"
+			 "Connection: close\r\n"
+			 "\r\n"
+			 "This is a response body, like it or not.",
+			 0));
 
     return OK;
 }
 
+/* Test read-til-EOF behaviour with SSL. */
+static int simple_eof(void)
+{
+    ne_session *sess = DEFSESS;
+
+    CALL(any_ssl_request(sess, serve_eof, SERVER_CERT, CA_CERT, NULL, NULL));
+    ne_session_destroy(sess);
+    return OK;
+}
+
+static int empty_truncated_eof(void)
+{
+    ne_session *sess = DEFSESS;
+    
+    CALL(any_ssl_request(sess, serve_response_unclean,
+			 "HTTP/1.0 200 OK\r\n" "\r\n",
+			 CA_CERT, NULL, NULL));
+
+    ne_session_destroy(sess);
+    return OK;
+}
+
+static int fail_truncated_eof(void)
+{
+    ne_session *sess = DEFSESS;
+    int ret;
+    
+    ONN("failed to load CA cert", ne_ssl_load_ca(sess, CA_CERT));
+    CALL(spawn_server(7777, serve_response_unclean,
+		      "HTTP/1.0 200 OK\r\n" "\r\n"
+		      "This is some content\n"
+		      "Followed by a truncation attack!\n"));
+    
+    ret = any_request(sess, "/foo");
+    CALL(await_server());
+    
+    ONV(ret != NE_ERROR,
+	("request failed with %d not error: `%s'", ret, ne_get_error(sess)));
+    ne_session_destroy(sess);
+    return OK;
+}
+    
+
 static int wildcard_match(void)
 {
     ne_session *sess;
@@ -308,6 +391,50 @@ static int wildcard_match(void)
     return OK;
 }
 
+/* Check that hostname comparisons are not cases-sensitive. */
+static int caseless_match(void)
+{
+    return accept_signed_cert("caseless.cert");
+}
+
+/* Test that the subjectAltName extension has precedence over the
+ * commonName attribute */
+static int subject_altname(void)
+{
+    return accept_signed_cert("altname.cert");
+}
+
+/* tests for multiple altNames. */
+static int two_subject_altname(void)
+{
+    return accept_signed_cert("altname2.cert");
+}
+
+static int two_subject_altname2(void)
+{
+    return accept_signed_cert("altname3.cert");
+}
+
+/* Test that a subject altname with *only* an eMail entry is
+ * ignored, and the commonName is used instead. */
+static int notdns_altname(void)
+{
+    return accept_signed_cert("altname4.cert");
+}
+
+/* test that the *most specific* commonName attribute is used. */
+static int multi_commonName(void)
+{
+    return accept_signed_cert("twocn.cert");
+}
+
+/* regression test for neon <= 0.23.4 where if commonName was the first
+ * RDN in the subject DN, it was ignored. */
+static int commonName_first(void)
+{
+    return accept_signed_cert("cnfirst.cert");
+}
+
 /* Verify callback which checks that the certificate has the predetermined
  * values (as per makekeys.sh). */
 static int check_DNs(void *userdata, int fs, const ne_ssl_certificate *cert)
@@ -391,7 +518,7 @@ static int no_verify(void)
     int count = 0;
 
     CALL(any_ssl_request(sess, serve_ssl, SERVER_CERT, CA_CERT, count_vfy,
-			 count_vfy));
+			 &count));
 
     ONN("verify callback called unnecessarily", count != 0);
 
@@ -445,6 +572,9 @@ static int fail_ssl_request(char *cert, char *cacert,
 	("no error in verification callback; request failed: %s",
 	 ne_get_error(sess)));
 
+    ONV(gotf & ~NE_SSL_FAILMASK,
+	("verification flags %x outside mask %x", gotf, NE_SSL_FAILMASK));
+
     /* check the failure flags were as expected. */
     ONV(failures != gotf,
 	("verification flags were %d not %d", gotf, failures));
@@ -464,25 +594,22 @@ static int fail_ssl_request(char *cert, char *cacert,
  * flagged as such. */
 static int fail_wrongCN(void)
 {
-    char *c;
-    CONCAT2(c, srcdir, "/wrongcn.pem");
-    return fail_ssl_request(c, c, "certificate with incorrect CN was accepted",
+    return fail_ssl_request("wrongcn.pem", "wrongcn.pem",
+			    "certificate with incorrect CN was accepted",
 			    NE_SSL_CNMISMATCH);
 }
 
 /* Check that an expired certificate is flagged as such. */
 static int fail_expired(void)
 {
-    char *c;
-    CONCAT2(c, srcdir, "/expired.pem");
+    char *c = ne_concat(srcdir, "/expired.pem", NULL);
     return fail_ssl_request(c, c, "expired certificate was accepted",
 			    NE_SSL_EXPIRED);
 }
 
 static int fail_notvalid(void)
 {
-    char *c;
-    CONCAT2(c, srcdir, "/notvalid.pem");
+    char *c = ne_concat(srcdir, "/notvalid.pem", NULL);
     return fail_ssl_request(c, c, "not yet valid certificate was accepted",
 			    NE_SSL_NOTYETVALID);
     
@@ -510,7 +637,7 @@ struct scache_args {
 };
 
 /* FIXME: factor out shared code with serve_ssl */
-static int serve_scache(nsocket *sock, void *ud)
+static int serve_scache(ne_socket *sock, void *ud)
 {
     struct scache_args *args = ud;
     SSL_SESSION *sess;
@@ -527,7 +654,7 @@ static int serve_scache(nsocket *sock, void *ud)
 
     args->count++;
 
-    CALL(do_ssl_response(sock, args->ctx, &sess));
+    CALL(do_ssl_response(sock, args->ctx, &sess, DEF_RESP, 0));
 
     /* dump session to child.log for debugging. */
     SSL_SESSION_print_fp(ne_debug_stream, sess);
@@ -588,6 +715,21 @@ static int client_cert_pem(void)
     return OK;
 }
 
+static void ccert_provider(void *userdata, ne_session *sess,
+			  const ne_ssl_dname *dn)
+{
+    ne_ssl_load_pem(sess, "client.cert", "client.key");
+}
+
+static int client_cert_provided(void)
+{
+    ne_session *sess = DEFSESS;
+    ne_ssl_provide_ccert(sess, ccert_provider, NULL);
+    CALL(any_ssl_request(sess, serve_ccert, SERVER_CERT, CA_CERT, NULL, NULL));
+    ne_session_destroy(sess);
+    return OK;
+}
+
 static int client_cert_pkcs12(void)
 {
     ne_session *sess = DEFSESS;
@@ -604,14 +746,58 @@ static int client_cert_pkcs12(void)
     return OK;
 }
 
+static int serve_tunnel(ne_socket *sock, void *ud)
+{
+    CALL(discard_request(sock));
+    
+    SEND_STRING(sock, "HTTP/1.1 200 OK\r\nServer: Fish\r\n\r\n");
+    
+    return serve_ssl(sock, ud);
+}
+
+/* neon versions <= 0.21.2 segfault here because ne_sock_close would
+ * be called twice on the socket after the server cert verification
+ * fails. */
+static int fail_tunnel(void)
+{
+    ne_session *sess = ne_session_create("https", "example.com", 443);
+    ne_session_proxy(sess, "localhost", 7777);
+
+    ONN("server cert verification didn't fail",
+	any_ssl_request(sess, serve_tunnel, SERVER_CERT, CA_CERT,
+			NULL, NULL) != NE_ERROR);
+    
+    ne_session_destroy(sess);
+    return OK;
+}
+
+static int proxy_tunnel(void)
+{
+    ne_session *sess = ne_session_create("https", "localhost", 443);
+    ne_session_proxy(sess, "localhost", 7777);
+    
+    /* CA cert is trusted, so no verify callback should be needed. */
+    CALL(any_ssl_request(sess, serve_tunnel, SERVER_CERT, CA_CERT,
+			 NULL, NULL));
+
+    ne_session_destroy(sess);
+    return OK;
+}
 
 /* TODO: code paths still to test in cert verification:
- * - server cert changes between connections
+ * - server cert changes between connections: Mozilla gives
+ * a "bad MAC decode" error for this; can do better?
  * - server presents no certificate (using ADH ciphers)... can
  * only really happen if they mess with the SSL_CTX and enable
  * ADH cipher manually; but good to check the failure case is 
  * safe.
+ * From the SSL book:
+ * - an early FIN should be returned as a possible truncation attack,
+ * NOT just an NE_SOCK_CLOSED.
+ * - unexpected close_notify is an error but not an attack.
+ * - never attempt session resumption after any aborted connection.
  */
+
 ne_test tests[] = {
     T(init),
 
@@ -619,20 +805,30 @@ ne_test tests[] = {
 
     T(load_pem_ccert),
     T(load_pkcs12_ccert),
-#if 0
-    /* somehow this test screws up later tests. */
     T(fail_load_ccerts),
-#endif
 
     T(simple),
+    T(simple_eof),
+    T(empty_truncated_eof),
+    T(fail_truncated_eof),
 
     T(client_cert_pem),
     T(client_cert_pkcs12),
+    T(client_cert_provided),
 
     T(parse_cert),
     T(no_verify),
     T(cache_verify),
     T(wildcard_match),
+    T(caseless_match),
+
+    T(subject_altname),
+    T(two_subject_altname),
+    T(two_subject_altname2),
+    T(notdns_altname),
+
+    T(multi_commonName),
+    T(commonName_first),
 
     T(fail_wrongCN),
     T(fail_expired),
@@ -641,6 +837,9 @@ ne_test tests[] = {
     T(fail_self_signed),
 
     T(session_cache),
+	
+    T(fail_tunnel),
+    T(proxy_tunnel),
 
     T(NULL) 
 };
