@@ -31,8 +31,8 @@
 #include "ne_md5.h"
 #include "ne_alloc.h"
 #include "ne_dates.h"
+#include "ne_string.h"
 
-#include "base64.h"
 #include "tests.h"
 
 static const struct {
@@ -54,6 +54,11 @@ static const struct {
     { "HTTP/1.1   200   OK", 1, 1, 200, "OK" },
     { "   HTTP/1.1 200 OK", 1, 1, 200, "OK" },
     { "Norman is a dog HTTP/1.1 200 OK", 1, 1, 200, "OK" },
+    
+    /* test that status-lines are cleansed. */
+    { "HTTP/1.1 200 Of\01Dubious\06Character", 1, 1, 200, 
+      "Of Dubious Character" },
+    
     { NULL }
 };
 
@@ -61,6 +66,8 @@ static const char *bad_sl[] = {
     "",
     "HTTP/1.1 1000 OK",
     "HTTP/1.1 1000",
+    "HTTP/-1.1 100 OK",
+    "HTTP/1.1 -100 OK",
     "HTTP/ 200 OK",
     "HTTP/",
     NULL
@@ -126,7 +133,8 @@ static int base64(void)
 {
 #define B64B(x, l, y) \
 do { char *_b = ne_base64(x, l); \
-ONV(strcmp(_b, y), ("B64(\"%s\") gave \"%s\"", x, _b)); free(_b); } while (0)
+ONV(strcmp(_b, y), ("B64(\"%s\") gave \"%s\" not \"%s\"", x, _b, y)); \
+free(_b); } while (0)
 #define B64(x, y) B64B(x, strlen(x), y)
 
     /* invent these with 
@@ -136,6 +144,7 @@ ONV(strcmp(_b, y), ("B64(\"%s\") gave \"%s\"", x, _b)); free(_b); } while (0)
     B64("bb", "YmI=");
     B64("ccc", "Y2Nj");
     B64("Hello, world", "SGVsbG8sIHdvcmxk");
+    B64("Aladdin:open sesame", "QWxhZGRpbjpvcGVuIHNlc2FtZQ==");
     B64("I once saw a dog called norman.\n", 
 	"SSBvbmNlIHNhdyBhIGRvZyBjYWxsZWQgbm9ybWFuLgo=");
     B64("The quick brown fox jumped over the lazy dog", 
@@ -195,11 +204,32 @@ static int parse_dates(void)
     return OK;
 }
 
+static int versioning(void)
+{
+#define GOOD(n,m,msg) ONV(ne_version_match(n,m), \
+("match of " msg " failed (%d.%d)", n, m))
+#define BAD(n,m,msg) ONV(ne_version_match(n,m) == 0, \
+("match of " msg " succeeded (%d.%d)", n, m))
+    GOOD(NEON_VERSION_MAJOR, NEON_VERSION_MINOR, "current version");
+    BAD(NEON_VERSION_MAJOR, NEON_VERSION_MINOR + 1, "later minor");
+    BAD(NEON_VERSION_MAJOR + 1, 0, "later major");
+#if NEON_VERSION_MINOR > 0
+    GOOD(NEON_VERSION_MAJOR, NEON_VERSION_MINOR - 1, "earlier minor");
+#endif
+#if NEON_VERSION_MAJOR > 0
+    BAD(NEON_VERSION_MAJOR - 1, 0, "earlier major");
+#endif
+#undef GOOD
+#undef BAD
+    return OK;
+}
+
 ne_test tests[] = {
     T(status_lines),
     T(md5),
     T(md5_alignment),
     T(base64),
     T(parse_dates),
+    T(versioning),
     T(NULL)
 };
